(*
 * <COPYRIGHT>
 * Copyright 2002 Joachim Schrod Network and Publication Consultance GmbH, Gerd Stolpmann
 *
 * <GPL>
 * This file is part of WDialog.
 *
 * WDialog is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * WDialog is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WDialog; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: wd_var_functions.ml,v 1.3 2003/02/16 01:07:03 stolpmann Exp $
 * ----------------------------------------------------------------------
 *
 *)

open Wd_types

(* --------------- Unicode string functions ---------------------- *)

(* These should be moved to ocamlnet! *)

let uni_length_of_string enc ?(byte_pos = 0) ?byte_len s =
  match enc with
      `Enc_utf8 ->
	let n = ref 0 in
	let k = ref byte_pos in
	let m = 
	  match byte_len with
	      None   -> String.length s
	    | Some l -> byte_pos + l
	in
	while !k < m do
	  incr n;
	  match s.[ !k ] with
	      ('\000'..'\127') ->
		incr k
	    | ('\128'..'\223') ->
		k := !k + 2
	    | ('\224'..'\239') ->
		k := !k + 3
	    | ('\240'..'\247') ->
		k := !k + 3
	    | _ ->
		assert false
	done;
	assert(!k = m);
	!n
    | (`Enc_utf16_le | `Enc_utf16_be | `Enc_utf16 ) ->
	assert false  (* UTF16 not supported by wdialog *)
    | `Enc_java ->
	assert false  (* Not supported by Wdialog *)
    | _ ->
	assert(Netconversion.is_single_byte enc);
	let m = 
	  match byte_len with
	      None   -> String.length s
	    | Some l -> l
	in
	m
;;


let uni_array_of_string enc s =
  match enc with
      `Enc_utf8 ->
	let l = uni_length_of_string enc s in
	let a = Array.make l 0 in
	let k = ref 0 in
	let dummy_buf = String.create 0 in
	ignore(Netconversion.recode 
		 ~in_enc:enc 
		 ~in_buf:s
		 ~in_pos:0
		 ~in_len:(String.length s)
		 ~out_enc:(`Enc_subset(enc, fun _ -> false))
		 ~out_buf:dummy_buf
		 ~out_pos:0
		 ~out_len:0
		 ~max_chars:l
		 ~subst:(fun x -> a.( !k ) <- x; incr k; ""));
	assert(!k = l);
	a
    | (`Enc_utf16_le | `Enc_utf16_be | `Enc_utf16 ) ->
	assert false  (* UTF16 not supported by wdialog *)
    | `Enc_java ->
	assert false  (* Not supported by Wdialog *)
    | (`Enc_usascii | `Enc_iso88591) ->
	let l = String.length s in
	let a = Array.make l 0 in
	for k = 0 to l - 1 do a.(k) <- Char.code s.[k] done;
	a
    | _ ->
	assert(Netconversion.is_single_byte enc);
	let conv = Netconversion.to_unicode ~in_enc:enc in
	let l = String.length s in
	let a = Array.make l 0 in
	for k = 0 to l - 1 do 
	  let x = Char.code s.[k] in
	  let x' = conv x in
	  a.(k) <- x'
	done;
	a
;;


let string_of_uni_array enc a =
    match enc with
      `Enc_utf8 ->
	let mk = Netconversion.makechar enc in
	let a' = Array.map mk a in
	let n = ref 0 in
	for k = 0 to Array.length a' - 1 do
	  n := !n + String.length a'.(k)
	done;
	let s = String.create !n in
	n := 0;
	for k = 0 to Array.length a' - 1 do
	  let t = a'.(k) in
	  String.blit t 0 s !n (String.length t);
	  n := !n + String.length t
	done;
	s
    | (`Enc_utf16_le | `Enc_utf16_be | `Enc_utf16 ) ->
	assert false  (* UTF16 not supported by wdialog *)
    | `Enc_java ->
	assert false  (* Not supported by Wdialog *)
    | (`Enc_usascii | `Enc_iso88591) ->
	let l = Array.length a in
	let s = String.create l in
	for k = 0 to l - 1 do
	  s.[ k ] <- Char.chr a.(k)
	done;
	s
    | _ ->
	assert(Netconversion.is_single_byte enc);
	let mk = Netconversion.makechar enc in
	let l = Array.length a in
	let s = String.create l in
	for k = 0 to l - 1 do 
	  let u = mk a.(k) in
	  s.[ k ] <- u.[ 0 ]
	done;
	s
;;

(* --------------------------------------------------------------- *)

let size (dlg : dialog_type) args =
  match args with
      [a1] ->
	(match a1 with
	     String_value s -> 
	       let enc = (dlg#application#charset :> Pxp_types.encoding) in
	       let n = uni_length_of_string enc s in
	       String_value (string_of_int n)
	   | Enum_value e ->
	       String_value (string_of_int (List.length e))
	   | Dyn_enum_value d ->
	       String_value (string_of_int (List.length d))
	   | Alist_value a ->
	       String_value (string_of_int (List.length a))
	   | Dialog_value _ ->
	       failwith "function `size': not defined for dialogs"
	)
    | _ ->
	failwith "function `size': expects exactly one argument"
;;


let binary_int_operation name op dlg args =
  let rec compute args =
    match args with
	[] ->
	  failwith ("function `" ^ name ^ "': expects at least one argument")
      | [ r ] ->
	  r
      | a1 :: a2 :: args' ->
	  ( match (a1,a2) with
		(String_value s1,String_value s2) ->
		  let n1 = int_of_string s1 in
		  let n2 = int_of_string s2 in
		  let r = String_value(string_of_int(op n1 n2)) in
		  compute (r :: args')
	      | _ ->
		  failwith ("function `" ^ name ^ "': operand has bad type")
	  )
  in
  compute args
;;


let add =
  binary_int_operation "add" ( + ) ;;

let sub =
  binary_int_operation "sub" ( - ) ;;

let mul =
  binary_int_operation "mul" ( * ) ;;

let div =
  let divide n1 n2 =
    if n2 = 0 then failwith "Division by zero" else n1/n2 in
  binary_int_operation "div" divide ;;

let alist_operation name op dlg args =
  match args with
      [a1; a2] ->
	( match (a1,a2) with
	      (Alist_value alist, String_value s) ->
		op alist s
	    | _ ->
		failwith ("function `" ^ name ^ "': bad type of operands")
	)
    | _ ->
	failwith ("function `" ^ name ^ "': expects exactly two arguments")
;;

let assoc =
  alist_operation "assoc"
    (fun alist s ->
       try List.assoc s alist
       with Not_found ->
	 failwith "function `assoc': element not found")
;;

let nth =
  alist_operation "nth"
    (fun alist s ->
       try snd(List.nth alist (int_of_string s))
       with Failure _ ->
	 failwith "function `nth': element not found")
;;

let enum_operation name op dlg args =
  match args with
      [a1; a2] ->
	( match (a1,a2) with
	      (Dyn_enum_value de, String_value s) ->
		op de s
	    | (String_value t, String_value s) ->
		( try
		    let enum = dlg # declaration # enumeration t in
		    op enum.enum_definition s
		  with
		      Not_found ->
			failwith ("function `" ^ name ^ "': enumeration not found")
		)
	    | _ ->
		failwith ("function `" ^ name ^ "': bad type of operands")
	)
    | _ ->
	failwith ("function `" ^ name ^ "': expects exactly two arguments")
;;

let translate =
  enum_operation "translate"
    (fun dynenum s ->
       try String_value (List.assoc s dynenum)
       with Not_found ->
	 failwith "function `translate': element not found")
;;

let rev_translate =
  enum_operation "rev_translate"
    (fun dynenum s ->
       let dynenum' = List.map (fun (l,r) -> (r,l)) dynenum in
       try String_value (List.assoc s dynenum')
       with Not_found ->
	 failwith "function `rev_translate': element not found")
;;

let substring (dlg : dialog_type) args =
  let enc = (dlg#application#charset :> Pxp_types.encoding) in
  let compute s start length =
    let (k1,k2) = 
      if length >= 0 then (start,start+length) else (start+length,start) in
    let n1 = max 0 k1 in
    let n2 = min (uni_length_of_string enc s) k2 in
    let a = uni_array_of_string enc s in
    let a' = Array.sub a n1 (n2-n1) in
    string_of_uni_array enc a'
  in
  match args with
      [a1;a2] ->
	( match (a1,a2) with
	      (String_value s1,String_value s2) ->
		String_value(compute s1 (int_of_string s2) (String.length s1))
	    | _ ->
		failwith "function `substring': bad type of operand"
	)
    | [a1;a2;a3] ->
	( match (a1,a2,a3) with
	      (String_value s1,String_value s2,String_value s3) ->
		String_value(compute s1 (int_of_string s2) (int_of_string s3))
	    | _ ->
		failwith "function `substring': bad type of operand"
	)
    | _ ->
	failwith "function `substring': expects two or three arguments"
;;

let concat dlg args =
  let strings =
    List.map
      (function
	   String_value s -> s
	 | _ ->
	     failwith "function `concat': bad type of operand")
      args 
  in
  String_value(String.concat "" strings)
;;

let unary_string_operation name op dlg args =
  match args with
      [a1] ->
	( match a1 with
	      (String_value s) ->
		op dlg s
	    | _ ->
		failwith ("function `" ^ name ^ "': bad type of operands")
	)
    | _ ->
	failwith ("function `" ^ name ^ "': expects exactly one argument")
;;


let rec count_height s n k =
  let next_cr_or_lf = Pxp_lib.crlf_index_from s k in
  if next_cr_or_lf >= 0 then begin
    match s.[next_cr_or_lf] with
        '\010' ->
          count_height s (n+1) (next_cr_or_lf+1)
      | '\013' ->
          let l = String.length s in
          if (next_cr_or_lf+1 < l && s.[next_cr_or_lf+1] = '\010') then
            count_height s (n+1) (next_cr_or_lf+2)
          else
            count_height s (n+1) (next_cr_or_lf+1)
      | _ ->
          assert false
  end
  else n
;;


let height =
  unary_string_operation "height"
    (fun dlg s ->
       String_value(string_of_int(count_height s 1 0)))
;;


let rec count_width enc s w k =
  let next_cr_or_lf = Pxp_lib.crlf_index_from s k in
  if next_cr_or_lf >= 0 then begin
    match s.[next_cr_or_lf] with
        '\010' ->
	  let n = uni_length_of_string 
		    enc ~byte_pos:k ~byte_len:(next_cr_or_lf - k) s in
	  let w' = max w n in
          count_width enc s w' (next_cr_or_lf+1)
      | '\013' ->
	  let n = uni_length_of_string 
		    enc ~byte_pos:k ~byte_len:(next_cr_or_lf - k) s in
	  let w' = max w n in
          let l = String.length s in
          if (next_cr_or_lf+1 < l && s.[next_cr_or_lf+1] = '\010') then
            count_width enc s w' (next_cr_or_lf+2)
          else
            count_width enc s w' (next_cr_or_lf+1)
      | _ ->
          assert false
  end
  else w
;;


let width =
  unary_string_operation "width"
    (fun (dlg : dialog_type) s ->
       let enc = (dlg#application#charset :> Pxp_types.encoding) in
       String_value(string_of_int(count_width enc s 0 0)))
;;


let var =
  unary_string_operation "var"
    (fun dlg s ->
       try
	 dlg # variable s
       with
	   No_such_variable msg ->
	     failwith ("function `var': no such variable: `" ^ msg ^ "'")
    )
;;


let const_operation name op (dlg:dialog_type) args =
  if args <> [] then
    failwith ("function `" ^ name ^ "': expects exactly no argument");
  op dlg
;;


let dialog =
  const_operation "dialog"
    (fun dlg -> String_value(dlg # name))
;;

let page =
  const_operation "page"
    (fun dlg -> String_value(dlg # page_name))
;;

let language =
  const_operation "language"
    (fun dlg ->
       match dlg # declaration # language_variable with
	   Some v ->
	     ( try
		 dlg # variable v
	       with
		   No_such_variable msg ->
		     failwith ("function `language': no such variable: `" ^ msg ^ "'")
	     )
	 | None ->
	     String_value ""
    )
       


(* ======================================================================
 * History:
 * 
 * $Log: wd_var_functions.ml,v $
 * Revision 1.3  2003/02/16 01:07:03  stolpmann
 * 	size, substring, width: string positions are measured as
 * the number of characters, not as the number of bytes
 *
 * Revision 1.2  2002/11/03 20:56:56  stolpmann
 * 	New functions: translate, rev_translate
 *
 * Revision 1.1  2002/10/20 19:38:17  stolpmann
 * 	Initial revision
 *
 * 
 *)
